@@ -1,11 +1,13 @@
 import logging
+import os
 import re
 import time
 from collections import defaultdict
+from contextlib import contextmanager
 from functools import reduce
-from operator import and_
 from pathlib import Path
 from typing import Union, List, Tuple, Type, Dict, Set, Callable
+from uuid import uuid4
 
 import graphviz
 import networkx as nx
@@ -19,10 +21,10 @@ from tqdm import tqdm
 
 from .file import File, HDU
 from .graph import Graph
-from .hierarchy import Multiple, Hierarchy, Graphable, Optional, OneOf
+from .hierarchy import Multiple, Hierarchy, Graphable, OneOf
 from .path_finding import find_path
 from .readquery import Query
-from .readquery.digraph import partial_reverse
+from .readquery.exceptions import UserError, CardinalityError
 from .readquery.results import RowParser
 from .utilities import make_plural, make_singular
 from .writequery import Unwind
@@ -116,13 +118,13 @@ def expand_template_relation(relation):
     """
     Returns a list of relations that relate to each non-template class
     e.g.
-    >>> expand_template_relation(Multiple(L1StackSpectrum))
-    [Multiple(L1SingleSpectrum), Multiple(L1OBStackSpectrum), Multiple(L1SuperstackSpectrum)]
+    >>> expand_template_relation(Multiple(L1StackedSpectrum))
+    [Multiple(L1SingleSpectrum), Multiple(L1StackSpectrum), Multiple(L1SuperstackSpectrum)]
     """
     if not relation.node.is_template:
         return [relation]
     subclasses = [cls for cls in get_all_subclasses(relation.node) if not cls.is_template]
-    return [Multiple(subclass, 0, relation.maxnumber, relation.constrain, relation.relation_idname) for subclass in subclasses]
+    return [Multiple(subclass, 0, relation.maxnumber, relation.constrain, relation.relation_idname, relation.one2one) for subclass in subclasses]
 
 
 def add_relation_graph_edge(graph, parent, child, relation: Multiple):
@@ -142,7 +144,11 @@ def add_relation_graph_edge(graph, parent, child, relation: Multiple):
         relation.instantate_node()
         # only parent-->child is in the database
         relstyle = 'solid' if relation.maxnumber == 1 else 'dashed'
-        if child_defines_parents:  # i.e. parents = [...] is set in the class for this object
+        if parent is child:
+            for a, b in [(parent, child), (child, parent)]:
+                graph.add_edge(a, b, singular=relation.maxnumber == 1,
+                               optional=relation.minnumber == 0, style=relstyle)
+        elif child_defines_parents:  # i.e. parents = [...] is set in the class for this object
             # child instance has n of type Parent, parent instance has unknown number of type Child
             parent = relation.node  # reset from new relations
             graph.add_edge(child, parent, singular=relation.maxnumber == 1,
@@ -234,27 +240,30 @@ def plot_graph(G, fname, format):
 class Data:
     filetypes = []
 
-    def __init__(self, rootdir: Union[Path, str] = '/beegfs/car/weave/weaveio/',
-                 host: str = '127.0.0.1', port=7687, write=False, dbname='neo4j',
-                 password='weavepassword', user='weaveuser', verbose=False):
+    def __init__(self, rootdir: Union[Path, str] = None,
+                 host: str = None, port=None, dbname=None,
+                 password=None, user=None, verbose=False):
         if verbose:
             logging.basicConfig(level=logging.INFO)
-        self.host = host
-        self.port = port
-        self.write_allowed = write
-        self.dbname = dbname
-        self._graph = None
-        self.password = password
-        self.user = user
-        self.rootdir = Path(rootdir)
+        self.dbname = dbname or os.getenv('WEAVEIO_DB', 'neo4j')
+        self.host = host or os.getenv('WEAVEIO_HOST', '127.0.0.1')
+        self.port = port or os.getenv('WEAVEIO_PORT', 7687)
+        self.password = password or os.getenv('WEAVEIO_PASSWORD', 'weavepassword')
+        self.user = user or os.getenv('WEAVEIO_USER', 'weaveuser')
+        self.rootdir = Path(rootdir or os.getenv('WEAVEIO_ROOTDIR', '/beegfs/car/weave/weaveio/'))
+
+        self.write_allowed = False
         self.query = Query(self)
         self.rowparser = RowParser(self.rootdir)
         self.filelists = {}
         self.relation_graphs = []
         for i, f in enumerate(self.filetypes):
             self.relation_graphs.append(make_relation_graph(hierarchies_from_files(*self.filetypes[:i+1])))
-        self.hierarchies = hierarchies_from_files(*self.filetypes, templates=True)
-        self.hierarchies.update({hh for h in self.hierarchies  for hh in get_all_class_bases(h)})
+        if self.filetypes:
+            self.hierarchies = hierarchies_from_files(*self.filetypes, templates=True)
+            self.hierarchies.update({hh for h in self.hierarchies  for hh in get_all_class_bases(h)})
+        else:
+            self.hierarchies = set()
         self.class_hierarchies = {h.__name__: h for h in self.hierarchies}
         self.singular_hierarchies = {h.singular_name: h for h in self.hierarchies}  # type: Dict[str, Type[Hierarchy]]
         self.plural_hierarchies = {h.plural_name: h for h in self.hierarchies if h.plural_name != 'graphables'}
@@ -294,7 +303,7 @@ class Data:
                 expanded_dependencies.add(o)
         return {o.__name__ for o in parents if o in expanded_dependencies}
 
-    def _path_to_hierarchy(self, from_obj: Type[Hierarchy], to_obj:  Type[Hierarchy], singular: bool):
+    def _path_to_hierarchy(self, g:nx.DiGraph, from_obj: Type[Hierarchy], to_obj:  Type[Hierarchy], singular: bool):
         """
         When searching for a path, the target is either above or below the source in one direction only
         If the target is defined as a child by another, then the search is redefined as for the predecessors of that child
@@ -309,7 +318,6 @@ class Data:
             If more than one path is returned, throw an ambiguous path exception
 
         """
-        g = self.relation_graphs[-1]
         return find_path(g, from_obj, to_obj, singular)
 
     def parents_of_defined_child(self, potential_child: Type[Hierarchy]) -> Set[Type[Hierarchy]]:
@@ -326,15 +334,19 @@ class Data:
     def path_to_hierarchy(self, from_obj: str, to_obj: str, singular: bool, descriptor=None, return_objs=False):
         a, b = map(self.singular_name, [from_obj, to_obj])
         from_obj, to_obj = self.singular_hierarchies[a], self.singular_hierarchies[b]
+        for level, g in enumerate(self.relation_graphs):
+            if from_obj in g and to_obj in g:
+                break
+        else:
+            raise nx.NodeNotFound(f"One of {from_obj} or {to_obj} is not in the graph".format(from_obj, to_obj))
         try:
-            path = self._path_to_hierarchy(from_obj, to_obj, singular)
-            g = self.relation_graphs[-1]
+            path = self._path_to_hierarchy(g, from_obj, to_obj, singular)
             singular = all(g.edges[(a, b)]['singular'] for a, b in zip(path[:-1], path[1:]))
             forwards = ['relation' not in g.edges[edge] for edge in zip(path[:-1], path[1:])]
             arrows = make_arrows(path, [not f for f in forwards], descriptor)
             if return_objs:
-                return arrows, singular, path
-            return arrows, singular
+                return arrows, singular, level, path
+            return arrows, singular, level
         except nx.NetworkXNoPath:
             if not singular:
                 to = f"multiple `{self.plural_name(b)}`"
@@ -361,37 +373,44 @@ class Data:
     def all_multiple_links_to_hierarchy(self, hierarchy: Type[Hierarchy]) -> Set[Type[Hierarchy]]:
         return self.all_links_to_hierarchy(hierarchy, lambda g, e: not g.edges[e]['singular'])
 
-    def write(self, collision_manager='track&flag'):
+    def write_cypher(self, collision_manager='track&flag'):
         if self.write_allowed:
             return self.graph.write(collision_manager)
         raise IOError(f"You have not allowed write operations in this instance of data (write=False)")
 
+    @property
+    @contextmanager
+    def write(self):
+        self.write_allowed = True
+        yield self.graph
+        self.write_allowed = False
+
     def is_unique_factor(self, name):
         return len(self.factor_hierarchies[name]) == 1
 
     @property
     def graph(self):
-        if self._graph is None:
-            d = {}
-            if self.password is not None:
-                d['password'] = self.password
-            if self.user is not None:
-                d['user'] = self.user
-            self._graph = Graph(host=self.host, port=self.port, name=self.dbname, write=self.write, **d)
-        return self._graph
+        d = {}
+        if self.password is not None:
+            d['password'] = self.password
+        if self.user is not None:
+            d['user'] = self.user
+        return Graph(host=self.host, port=self.port, name=self.dbname, write=self.write_allowed, **d)
 
     def make_constraints_cypher(self):
-        return {hierarchy: hierarchy.make_schema() for hierarchy in self.hierarchies}
+        d = {hierarchy: hierarchy.make_schema() for hierarchy in self.hierarchies}
+        d['temporary'] = ['CREATE CONSTRAINT ON (t:TemporaryMerge) ASSERT t.id IS UNIQUE']
+        return d
 
     def apply_constraints(self):
         if not self.write_allowed:
             raise IOError(f"Writing is not allowed")
         templates = []
         equivalencies = []
-        for hier, q in tqdm(self.make_constraints_cypher().items(), desc='applying constraints'):
-            if q is None:
+        for hier, qs in tqdm(self.make_constraints_cypher().items(), desc='applying constraints'):
+            if not qs:
                 templates.append(hier)
-            else:
+            for q in qs:
                 try:
                     self.graph.neograph.run(q)
                 except py2neo.ClientError as e:
@@ -399,19 +418,36 @@ class Data:
                        equivalencies.append(hier)
                        templates.append(hier)
         if len(templates):
-            print(f'No index/constraint was made for {templates}')
+            logging.info(f'No index/constraint was made for {templates}')
         if len(equivalencies):
-            print(f'EquivalentSchemaRuleAlreadyExists for {equivalencies}')
+            logging.info(f'EquivalentSchemaRuleAlreadyExists for {equivalencies}')
 
-    def drop_all_constraints(self):
+    def drop_all_constraints(self, indexes=True):
         if not self.write_allowed:
             raise IOError(f"Writing is not allowed")
         constraints = self.graph.neograph.run('CALL db.constraints() YIELD name return "DROP CONSTRAINT " + name + ";"')
         for constraint in tqdm(constraints, desc='dropping constraints'):
             self.graph.neograph.run(str(constraint)[1:-1])
+        if indexes:
+            self.drop_all_indexes()
+
+    def drop_all_indexes(self):
+        if not self.write_allowed:
+            raise IOError(f"Writing is not allowed")
+        indexes = self.graph.neograph.run('CALL db.indexes() yield name, labelsOrTypes where size(labelsOrTypes) > 0 return "DROP INDEX " + name + ";"')
+        for index in tqdm(indexes, desc='dropping indexes'):
+            self.graph.neograph.run(str(index)[1:-1])
+
+    def get_extant_files(self, *filetypes, return_only_fname=True):
+        filetypes = [self.singular_hierarchies[self.singular_name(f)].__name__ for f in filetypes]
+        if not filetypes:
+            filetypes = ['File']
+        r = 'fname' if return_only_fname else 'path'
+        result = self.graph.execute(f"MATCH (f:File) where any(l in labels(f) where l in $filetypes) RETURN DISTINCT f.{r}", filetypes=filetypes).to_series(dtype=str).values.tolist()
+        if return_only_fname:
+            return result
+        return list(map(lambda x: self.rootdir / x, result))
 
-    def get_extant_files(self):
-        return self.graph.execute("MATCH (f:File) RETURN DISTINCT f.fname").to_series(dtype=str).values.tolist()
 
     def raise_collisions(self):
         """
@@ -421,9 +457,24 @@ class Data:
         rel_collisions = self.graph.execute("MATCH ()-[c: _Collision]-() return c { .*}").to_data_frame()
         return node_collisions, rel_collisions
 
-    def read_files(self, *paths: Union[Path, str], raise_on_duplicate_file=False,
-                   collision_manager='ignore', batch_size=None, halt_on_error=True,
-                   dryrun=False, do_not_apply_constraints=False) -> pd.DataFrame:
+    def get_state_of_file(self, fname: str) -> int:
+        """Returns the state (timestamp) of the creation of the last data product belong to a given file"""
+        c = self.data.execute("match (f: File {fname: $fname}) optional match (f)<-[:is_required_by]-(h:Hierarchy) "
+                          "where not h:HDU with  h.`_dbcreated` as d order by d desc return d limit 1", fname=fname)
+        return c.evaluate()
+
+    def restore_state(self, state: int):
+        """
+        Restores database to the given state (timestamp). Danger zone here.
+        """
+        if not self.write_allowed:
+            raise IOError(f"Writing is not allowed")
+        return self.graph.execute('match (n) where n.`_dbcreated` > $timestamp detach delete n', timestamp=state)
+
+    def write_files(self, *paths: Union[Path, str], raise_on_duplicate_file=False,
+                    collision_manager='ignore', batch_size=None, parts=None, halt_on_error=True,
+                    dryrun=False, do_not_apply_constraints=False, test_one=False,
+                    debug=False, debug_time=False) -> pd.DataFrame:
         """
         Read in the files given in `paths` to the database.
         `collision_manager` is the method with which the database deals with overwriting data.
@@ -444,34 +495,66 @@ class Data:
                 raise ValueError(f"{path} matches more than 1 file type: {matches} with `{[m.match_pattern for m in matches]}`")
             filetype = matches[0]
             filetype_batch_size = filetype.recommended_batchsize if batch_size is None else batch_size
-            slices = filetype.get_batches(path, filetype_batch_size)
-            batches += [(filetype, path.relative_to(self.rootdir), slc) for slc in slices]
+            slices = filetype.get_batches(path, filetype_batch_size, parts)
+            batches += [(filetype, path.relative_to(self.rootdir), slc, part) for slc, part in slices]
         elapsed_times = []
         stats = []
         timestamps = []
         if dryrun:
             logging.info(f"Dryrun: will not write to database. However, reading is permitted")
-        bar = tqdm(batches)
-        for filetype, fname, slc in bar:
-            bar.set_description(f'{fname}[{slc.start}:{slc.stop}]')
+        if test_one:
+            batches = batches[:1]
+        _, path, slc, part = batches[0]
+        bar = tqdm(batches, desc=f'{path}[{slc.start}:{slc.stop}:{part}]')
+        if debug or debug_time:
+            with open('debug-timestamp.log', 'w') as f:
+                pass
+        for filetype, path, slc, part in bar:
+            bar.set_description(f'{path}[{slc.start}:{slc.stop}:{part}]')
             try:
                 if raise_on_duplicate_file:
-                    if len(self.graph.execute('MATCH (f:File {fname: $fname})', fname=fname)) != 0:
-                        raise FileExistsError(f"{fname} exists in the DB and raise_on_duplicate_file=True")
-                with self.write(collision_manager) as query:
-                    filetype.read(self.rootdir, fname, slc)
+                    if len(self.graph.execute('MATCH (f:File {fname: $fname})', fname=path.name)) != 0:
+                        raise FileExistsError(f"{path.name} exists in the DB and raise_on_duplicate_file=True")
+                with self.write_cypher(collision_manager) as query:
+                    filetype.read(self.rootdir, path, slc, part)
                 cypher, params = query.render_query()
+                uuid = f"//{uuid4()}"
+                cypher = '\n'.join([uuid, 'CYPHER runtime=interpreted', cypher])  # runtime is to avoid neo4j bug with pipelines: https://github.com/neo4j/neo4j/issues/12441
+                if debug:
+                    with open('debug-query.log', 'w') as f:
+                        f.write(cypher)
+                    with open('debug-params.log', 'w') as f:
+                        f.write(self.graph.output_for_debug(**params, silent=True))
                 start = time.time()
                 if not dryrun:
-                    results = self.graph.execute(cypher, **params)
-                    stats.append(results.stats())
-                    timestamp = results.evaluate()
+                    try:
+                        results = self.graph.execute(cypher, **params)
+                        stats.append(results.stats())
+                        timestamp = results.evaluate()
+                    except (ConnectionError, RuntimeError, IndexError, ConnectionResetError) as e:
+                        is_running = True
+                        while is_running:
+                            is_running = self.graph.execute("CALL dbms.listQueries() YIELD query WHERE query STARTS WITH $uuid return count(*)", uuid=uuid).evaluate()
+                            time.sleep(1)
+                            bar.set_description(f'{path}[{slc.start}:{slc.stop}:{part}] (waiting for query to finish)')
+                        r = self.graph.execute('MATCH (f:File {fname: $fname}) return timestamp()', fname=path.name)
+                        successful = r.evaluate()
+                        timestamp = successful
+                        if not successful:
+                            raise ConnectionError(f"{path} could not be written to the database see neo4j logs for more details") from e
+                        stats.append(r.stats())
                     if timestamp is None:
-                        logging.warning(f"This query terminated early due to an empty input table/data. "
-                             f"Adjust your `.read` method to allow for empty tables/data")
+                        logging.warning(f"This query terminated early due to either an empty input table/data or "
+                                        f"a match within the query returned no matches. "
+                                        f"Adjust your `.read` method and query to allow for empty tables/data")
                     timestamps.append(timestamp)
+                else:
+                    self.graph.execute('EXPLAIN\n' + 'CYPHER runtime=interpreted\n' + cypher, **params)
                 elapsed_times.append(time.time() - start)
-            except (ClientError, DatabaseError, FileExistsError) as e:
+                if debug or debug_time:
+                    with open('debug-timestamp.log', 'a') as f:
+                        f.write(str(elapsed_times[-1]) + '\n')
+            except (ClientError, DatabaseError, FileExistsError, ConnectionError) as e:
                 logging.exception('ClientError:', exc_info=True)
                 if halt_on_error:
                     raise e
@@ -480,14 +563,15 @@ class Data:
             df = pd.DataFrame(stats)
             df['timestamp'] = timestamps
             df['elapsed_time'] = elapsed_times
-            _, df['fname'], slcs = zip(*batches)
+            _, df['fname'], slcs, parts = zip(*batches)
             df['batch_start'], df['batch_end'] = zip(*[(i.start, i.stop) for i in slcs])
+            df['part'] = parts
         elif dryrun:
-            df = pd.DataFrame(columns=['elapsed_time', 'fname', 'batch_start', 'batch_end'])
+            df = pd.DataFrame(columns=['elapsed_time', 'fname', 'batch_start', 'batch_end', 'part'])
             df['elapsed_time'] = elapsed_times
         else:
             df = pd.DataFrame(columns=['timestamp', 'elapsed_time', 'fname', 'batch_start', 'batch_end'])
-        return df.set_index(['fname', 'batch_start', 'batch_end'])
+        return df.set_index(['fname', 'batch_start', 'batch_end', 'part'])
 
     def find_files(self, *filetype_names, skip_extant_files=True):
         filelist = []
@@ -502,19 +586,19 @@ class Data:
             filelist += [i for i in filetype.match_files(self.rootdir, self.graph)]
         if skip_extant_files:
             extant_fnames = self.get_extant_files() if skip_extant_files else []
-            filtered_filelist = [i for i in filelist if str(i.relative_to(self.rootdir)) not in extant_fnames]
+            filtered_filelist = [i for i in filelist if str(i.name) not in extant_fnames]
         else:
             filtered_filelist = filelist
         diff = len(filelist) - len(filtered_filelist)
         if diff:
             print(f'Skipping {diff} extant files (use skip_extant_files=False to go over them again)')
-        return filtered_filelist
+        return [f for f in tqdm(filtered_filelist, desc='getting MOS files') if File.check_mos(f)]
 
-    def read_directory(self, *filetype_names, collision_manager='ignore', skip_extant_files=True, halt_on_error=False,
-                        dryrun=False) -> pd.DataFrame:
+    def write_directory(self, *filetype_names, collision_manager='ignore', skip_extant_files=True, halt_on_error=False,
+                        batch_size=None, parts=None, dryrun=False) -> pd.DataFrame:
         filtered_filelist = self.find_files(*filetype_names, skip_extant_files=skip_extant_files)
-        return self.read_files(*filtered_filelist, collision_manager=collision_manager, halt_on_error=halt_on_error,
-                                dryrun=dryrun)
+        return self.write_files(*filtered_filelist, collision_manager=collision_manager, halt_on_error=halt_on_error,
+                                dryrun=dryrun, batch_size=batch_size, parts=parts)
 
     def _validate_one_required(self, hierarchy_name):
         hierarchy = self.singular_hierarchies[hierarchy_name]
@@ -637,7 +721,7 @@ class Data:
                     try:
                         return self.relative_names[name]
                     except KeyError:
-                        return self.singular_hierarchies[name].singular_name
+                        return self.singular_hierarchies[name].plural_name
         if '.' in name:
             pattern = name.lower().split('.')
             if any(map(self.is_plural_name, pattern)):
@@ -673,6 +757,10 @@ class Data:
         return make_singular(name)
 
     def is_valid_name(self, name):
+        if not isinstance(name, str):
+            return False
+        if self.is_plural_name(name) or self.is_singular_name(name):
+            return True
         if isinstance(name, str):
             pattern = name.split('.')
             if len(pattern) == 1:
@@ -685,6 +773,11 @@ class Data:
         Returns True if name is a plural name of a hierarchy
         e.g. spectra is plural for Spectrum
         """
+        if not isinstance(name, str):
+            return False
+        if name in self.plural_hierarchies or name in self.plural_factors or\
+                   name in self.plural_idnames or name in self.plural_relative_names:
+            return True
         pattern = name.split('.')
         if len(pattern) == 1:
             return name in self.plural_hierarchies or name in self.plural_factors or\
@@ -692,6 +785,11 @@ class Data:
         return all(self.is_plural_name(n) for n in pattern)
 
     def is_singular_name(self, name):
+        if not isinstance(name, str):
+            return False
+        if name in self.singular_hierarchies or name in self.singular_factors or \
+           name in self.singular_idnames or name in self.relative_names:
+            return True
         pattern = name.split('.')
         if len(pattern) == 1:
             return name in self.singular_hierarchies or name in self.singular_factors or \
@@ -699,10 +797,16 @@ class Data:
         return all(self.is_singular_name(n) for n in pattern)
 
     def __getitem__(self, address):
-        return self.query.__getitem__(address)
+        try:
+            return self.query.__getitem__(address)
+        except (UserError, KeyError) as e:
+            self.query.raise_error_with_suggestions(address, e)
 
     def __getattr__(self, item):
-        return self.query.__getattr__(item)
+        try:
+            return self.query.__getattr__(item)
+        except (UserError, KeyError) as e:
+            self.query.raise_error_with_suggestions(item, e)
 
     def plot_relations(self, i=-1, show_hdus=True, fname='relations', format='pdf', include=None):
         graph = self.relation_graphs[i]
@@ -718,6 +822,16 @@ class Data:
             G = nx.subgraph_view(G, lambda n: n in include_list)
         plot_graph(G, fname, format)
 
+    def find_names(self, guess, n=10):
+        """
+        Finds the closest name to guess in the hierarchy. Returns the top n matches.
+        """
+        sources = [self.singular_hierarchies, self.plural_hierarchies, self.singular_factors, self.plural_factors]
+        suggestions = {i for source in sources for i in source}
+        inorder = sorted(suggestions, key=lambda x: textdistance.jaro_winkler(guess, x), reverse=True)
+        return inorder[0:n]
+
+
     def _autosuggest(self, a, relative_to=None):
         a = self.singular_name(a)
         if relative_to is not None:
@@ -733,6 +847,8 @@ class Data:
                 except NetworkXNoPath:
                     hier = h.singular_name
                     factors = h.products_and_factors
+                except NodeNotFound:
+                    continue
                 else:
                     hier = h.plural_name
                     factors = [self.plural_name(f) for f in h.products_and_factors]
@@ -750,11 +866,11 @@ class Data:
         inorder = sorted(list(set(suggestions)), key=lambda x: distance(a, x), reverse=distance_reverse)
         return inorder[:3]
 
-    def autosuggest(self, a: str, relative_to: str = None, exception=None):
-        suffix = ''
-        try:
+    def autosuggest(self, a: str, relative_to: str = None, exception: Exception = None):
+        if isinstance(exception, CardinalityError):
+            l = [self.plural_name(a)]
+        else:
             l = self._autosuggest(a, relative_to)
-            string = '\n'.join([f'{i}. {s}' for i, s in enumerate(l, start=1)])
-        except ImportError:
-            raise AttributeError(f"`{a}` not understood.{suffix}") from exception
-        raise AttributeError(f"`{a}` not understood, did you mean one of:\n{string}{suffix}") from exception
\ No newline at end of file
+        string = '\n'.join([f'{i}. {s}' for i, s in enumerate(l, start=1)])
+        msg = f"did you mean one of:\n{string}"
+        raise exception.__class__(f"{str(exception.args[0])}\n{msg}") from exception
\ No newline at end of file
