@@ -1,9 +1,11 @@
+import logging
 from datetime import datetime
 from time import sleep
 from warnings import warn
 
 import py2neo
 from astropy.table import Table
+from pathlib import Path
 from py2neo import Graph as NeoGraph
 
 import numpy as np
@@ -12,11 +14,12 @@ from weaveio.context import ContextMeta
 from weaveio.writequery import CypherQuery
 
 missing_types = {int:  np.inf, float: np.inf, str: '<MISSING>', type(None): np.inf, None: np.inf, bool: np.inf, datetime: datetime(1900, 1, 1, 0, 0)}
-convert_types = {int:  float, float: float, str: str, type(None): float, None: float, bool: float,
+convert_types = {int:  int, float: float, str: str, type(None): float, None: float, bool: float,
                  datetime: lambda x: datetime(x.year, x.month, x.day, x.hour, x.minute, x.second),
                  list: list, tuple: tuple, np.ndarray: np.ndarray, pd.DataFrame: pd.DataFrame, pd.Series: pd.Series,
                  dict: dict, Table: Table,
-                 np.int64: int, np.float64: float, np.float32: float, np.int32:int}
+                 np.int64: int, np.float64: float, np.float32: float, np.int32:int,
+                 Path: str}
 
 def is_null(x):
     try:
@@ -40,7 +43,7 @@ def _convert_datatypes(x, nan2missing=True, none2missing=True, surrounding_type=
             r = tuple(r)
         return r
     elif isinstance(x, dict):
-        return {_convert_datatypes(k, nan2missing, none2missing): _convert_datatypes(v, nan2missing, none2missing) for k, v in x.items()}
+        return {str(_convert_datatypes(k, nan2missing, none2missing)): _convert_datatypes(v, nan2missing, none2missing) for k, v in x.items()}
     elif isinstance(x, (pd.DataFrame, pd.Series)):
         return _convert_datatypes(pd.DataFrame(x).reset_index().to_dict('records'), nan2missing, none2missing)
     elif isinstance(x, Table):
@@ -98,21 +101,23 @@ class Graph(metaclass=ContextMeta):
         return CypherQuery(collision_manager)
 
     def _execute(self, cypher, parameters, backoff=1, limit=10):
-        try:
-            return self.neograph.run(cypher, parameters=parameters)
-        except RuntimeError as e:
-            if backoff >= limit:
-                raise e
-            warn(f'Connection possibly busy, waiting {backoff} seconds to retry. Actual error was {e}')
-            sleep(backoff)
-            backoff *= 2
-            return self._execute(cypher, parameters, backoff, limit)
+        return self.neograph.auto(readonly=not self.write_allowed).run(cypher, parameters=parameters)
+        # try:
+        #     pass
+        # except (ConnectionError) as e:
+        #     logging.info(f'Connection possibly busy, waiting {backoff} seconds to retry. Actual error was {e}')
+        #     if backoff >= limit:
+        #         raise e
+        #     sleep(backoff)
+        #     backoff *= 2
+        # except (RuntimeError, IndexError) as e:
+        #     logging.info(f'Connection failed with an undefined py2neo error, waiting {backoff} seconds to retry. Actual error was {e}')
+        #     if backoff >= limit:
+        #         raise e
+        #     backoff *= 2  # dont sleep retry immediately
+        #     return self._execute(cypher, parameters, backoff, limit)
 
     def execute(self, cypher, **payload):
-        lower = cypher.lower()
-        if not self.write_allowed and ('create' in lower or 'merge' in lower or
-                                       'set' in lower or 'delete' in lower or 'detach' in lower):
-            raise IOError(f"Write is not allowed, set `write=True` to permit writing.")
         d = _convert_datatypes(payload, nan2missing=True, none2missing=True)
         try:
             return self._execute(cypher, d)
@@ -120,10 +125,11 @@ class Graph(metaclass=ContextMeta):
             raise ConnectionResetError(f"Py2neo dropped the connection because it was taking too long. "
                                        f"Split up your query using batch_size=??")
 
-    def output_for_debug(self, **payload):
+    def output_for_debug(self,  silent=False, **payload):
         d = _convert_datatypes(payload, nan2missing=True, none2missing=True)
-        warn(f"When parameters are output for debug in the neo4j desktop, it cannot be guaranteed the data types will remain the same. "
-             f"For certain, infs/nans/None are converted to strings (to avoid this, run your query without using `output_for_debug`)")
+        if not silent:
+            warn(f"When parameters are output for debug in the neo4j desktop, it cannot be guaranteed the data types will remain the same. "
+                 f"For certain, infs/nans/None are converted to strings (to avoid this, run your query without using `output_for_debug`)")
         return f':params {d}'
 
 Graph._context_class = Graph
