@@ -13,7 +13,8 @@ from typing import List, TYPE_CHECKING, Union, Tuple, Dict, Any
 
 from networkx import NetworkXNoPath
 
-from .base import BaseQuery, CardinalityError
+from .base import BaseQuery
+from .exceptions import CardinalityError, AttributeNameError, UserError
 from .parser import QueryGraph, ParserError
 from .utilities import is_regex, dtype_conversion
 from ..opr3.hierarchy import Exposure
@@ -21,13 +22,25 @@ from ..opr3.hierarchy import Exposure
 if TYPE_CHECKING:
     from weaveio.data import Data
 
+
 class GenericObjectQuery(BaseQuery):
     pass
 
 
-class PathError(SyntaxError):
-    pass
-
+def process_names(given_names, attrs: List['AttributeQuery']) -> List[str]:
+    for i, (given_name, attr) in enumerate(zip(given_names, attrs)):
+        if given_name is None:
+            given_names[i] = attr._factor_name
+    if len(set(given_names)) == len(given_names):
+        return given_names
+    names = [a._factor_name for a in attrs]
+    if len(set(names)) == len(names):
+        return names
+    for i, n in enumerate(names):
+        if names.count(n) > 1:
+            j = names[:i].count(n)
+            names[i] = f"{n}{j}"
+    return [a._factor_name for a in attrs]  # todo: this could be better
 
 class ObjectQuery(GenericObjectQuery):
     def _precompile(self) -> 'TableQuery':
@@ -37,12 +50,11 @@ class ObjectQuery(GenericObjectQuery):
         """
         Obj = self._data.class_hierarchies[self._obj]
         if Obj.idname is not None:
-            if len(Obj.products_and_factors) == 2:
-                attr = Obj.products_and_factors[0]  # take the only data in it
-            else:
-                attr = Obj.idname
+            attr = Obj.idname
         elif len(Obj.products_and_factors) == 1:
             attr = Obj.products_and_factors[0]  # take the only data in it
+        elif 'value' in Obj.products_and_factors:
+            attr = 'value'
         else:
             raise SyntaxError(f"{self._obj} cannot be returned/identified since it doesn't define any unique idname. "
                               f"If you want to return all singular data for {self._obj} use ...['*']")
@@ -83,15 +95,16 @@ class ObjectQuery(GenericObjectQuery):
         traversal, expands
         e.g. `ob.runs`  reads "for each ob, get its runs"
         """
+        plural = self._data.plural_name(obj)
+        err_msg = f"There is no singular {obj} relative to {self._obj} try using its plural {plural}"
         try:
-            path, single = self._get_path_to_object(obj, want_single)
+            path, single, _ = self._get_path_to_object(obj, want_single)
         except NetworkXNoPath:
-            if want_single:
-                plural = self._data.plural_name(obj)
-                msg = f"There is no singular {obj} relative to {self._obj} try using its plural {plural}"
-            else:
-                msg = f"There is no {obj} relative to {self._obj}"
-            raise CardinalityError(msg)
+            if not want_single:
+                err_msg = f"There is no {obj} relative to {self._obj}"
+            raise CardinalityError(err_msg)
+        if not single and want_single:
+            raise CardinalityError(err_msg)
         n = self._G.add_traversal(self._node, path, obj, single)
         return ObjectQuery._spawn(self, n, obj, single=want_single)
 
@@ -104,7 +117,7 @@ class ObjectQuery(GenericObjectQuery):
         e.g. `obs[1234]` filters to the single ob with obid=1234
         """
         param = self._G.add_parameter(index)
-        path, single = self._get_path_to_object(obj, False)
+        path, single, _ = self._get_path_to_object(obj, False)
         travel = self._G.add_traversal(self._node, path, obj, single)
         i = self._G.add_getitem(travel, 'id')
         eq, _ = self._G.add_scalar_operation(i, f'{{0}} = {param}', f'id={index}')
@@ -113,7 +126,7 @@ class ObjectQuery(GenericObjectQuery):
 
     def _traverse_by_object_indexes(self, obj, indexes: List):
         param = self._G.add_parameter(indexes)
-        path, single = self._get_path_to_object(obj, False)
+        path, single, _ = self._get_path_to_object(obj, False)
         one_id = self._G.add_unwind_parameter(self._node, param)
         travel = self._G.add_traversal(self._node, path, obj, single, one_id)
         i = self._G.add_getitem(travel, 'id')
@@ -157,28 +170,23 @@ class ObjectQuery(GenericObjectQuery):
         obj['factora', 'obj.factorb']
         """
         attrs = []
-        names = []
         for item in items:
             if isinstance(item, ObjectQuery):
                 item = item._precompile()
             if isinstance(item, AttributeQuery):
                 attrs.append(item)
-                names.append(f"{item._names[0]}")
             else:
                 new = self.__getitem__(item)
                 if isinstance(new, ObjectQuery):
                     new = new._precompile()
                 if isinstance(new, list):
                     for a in new:
-                        if a._factor_name not in names or a._factor_name is None:
-                            names.append(f"{a._names[0]}")
-                            attrs.append(a)
-                elif new._factor_name not in names or new._factor_name is None:
-                    names.append(f"{new._names[0]}")
-                    attrs.append(new)
+                        attrs.append(a)
+                attrs.append(new)
         force_plurals = [not a._single for a in attrs]
         is_products = [a._is_products[0] for a in attrs]
         n = self._G.add_results_table(self._node, [a._node for a in attrs], force_plurals, dropna=[self._node])
+        names = process_names([i if isinstance(i, str) else None for i in items], attrs)
         return TableQuery._spawn(self, n, names=names, is_products=is_products, attrs=attrs)
 
     def _traverse_to_relative_object(self, obj, index):
@@ -194,7 +202,7 @@ class ObjectQuery(GenericObjectQuery):
         """
         obj, obj_singular = self._normalise_object(obj)
         singular_name = self._data.singular_name(index)
-        path, single = self._get_path_to_object(obj, False)
+        path, single, _ = self._get_path_to_object(obj, False)
         if not single and singular_name == index:
             raise SyntaxError(f"Relative index `{index}` is plural relative to `{self._obj}`.")
         n = self._G.add_traversal(self._node, path, obj, False)
@@ -251,7 +259,7 @@ class ObjectQuery(GenericObjectQuery):
                         obj, attr = item.split('.')
                         return self.__getitem__(obj).__getitem__(attr)
                     except ValueError:
-                        raise ValueError(f"{item} cannot be parsed as an `obj.attribute`.")
+                        raise AttributeNameError(f"{item} cannot be parsed as an `obj.attribute`.")
                 try: # try assuming its an object
                     obj, single = self._normalise_object(item)
                     if obj == self._obj:
@@ -260,30 +268,24 @@ class ObjectQuery(GenericObjectQuery):
                 except (KeyError, ValueError):  # assume its an index of some kind
                     singular = self._data.singular_name(item)
                     if singular in self._data.relative_names:  # if it's a relative id
-                        if by_getitem:  # filter by relative relation: ob.runs['red'] gets red runs
-                            if singular != item:
-                                raise SyntaxError(f"Filtering by relative index `{item}` must use its singular name `{singular}`")
-                            print(self._data.relative_names[singular])
-                            if self._previous._obj not in self._data.relative_names[singular]:
-                                raise PathError(f"There are no `{singular}` `{self._obj}` of `{self._previous._obj}`")
-                            return self._previous._traverse_to_relative_object(self._obj, item)
-                        else:  # l1singlespectrum.adjunct gets the adjunct spectrum of a spectrum
-                            try:
-                                relation = self._data.relative_names[singular][self._obj]
-                            except KeyError:
-                                raise PathError(f"`{self._obj}` has no relative relation called `{singular}`")
-                            return self._traverse_to_relative_object(relation.node.__name__, item)
+                        # l1singlespectrum.adjunct gets the adjunct spectrum of a spectrum
+                        try:
+                            relation = self._data.relative_names[singular][self._obj]
+                        except KeyError:
+                            raise AttributeNameError(f"`{self._obj}` has no relative relation called `{singular}`")
+                        return self._traverse_to_relative_object(relation.node.__name__, item)
                     elif by_getitem:  # otherwise treat as an index
                         return self._previous._traverse_by_object_index(self._obj, item)
                     else:
-                        raise KeyError(f"Unknown attribute `{item}`")
+                        raise AttributeNameError(f"Unknown attribute `{item}`")
 
     def _getitem_handled(self, item, by_getitem):
         try:
             return self._getitem(item, by_getitem)
-        except (KeyError, ValueError, NetworkXNoPath, CardinalityError) as e:
-            # self._data.autosuggest(item, self._obj)
-            raise e
+        except UserError as e:
+            if isinstance(item, AttributeQuery):
+                raise e
+            self._data.autosuggest(item, self._obj, e)
 
     def __getattr__(self, item):
         return self._getitem_handled(item, False)
@@ -292,7 +294,17 @@ class ObjectQuery(GenericObjectQuery):
         return self._getitem_handled(item, True)
 
     def __eq__(self, other):
-        return self._select_attribute('id', True).__eq__(other)
+        obj = self._data.class_hierarchies[self._obj]
+        potential_id = obj.products_and_factors
+        if 'id' in potential_id:
+            i = 'id'
+        elif 'name' in potential_id:
+            i = 'name'
+        elif 'value' in potential_id:
+            i = 'value'
+        else:
+            raise AttributeNameError(f"`A {self._obj}` has no attribute `id`, `name`, or `value` to compare with. Choose a specific attribute like `{obj.singular_name}.id`.")
+        return self._select_attribute(i, True).__eq__(other)
 
 
 class Query(GenericObjectQuery):
